from PIL import Image
from glfw.GLFW import *

from avatar import *
from objects import myObjects
from shader import *

# sirka a vyska okna
scr_w: int = 800
scr_h: int = 600

# pozicia mysi
mouse_lastX: float = scr_w / 2.0
mouse_lastY: float = scr_h / 2.0
mouse_start: bool = True

# uplynuty cas medzi aktualnym a povodnym casovym okamihom
dTime: float = 0.0
lTime: float = 0.0

# avatar
myAvatar = Avatar(glm.vec3(0.0, 2.0, 10.0), scr_w / scr_h)


def main() -> int:
    global dTime, lTime

    # inicializacia glfw okna
    glfwInit()
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3)
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3)
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE)

    # vytvorenie glfw okna
    glfwWin = glfwCreateWindow(scr_w, scr_h, "PyGLSL v1.0", None, None)
    if glfwWin is None:
        print("Nepodarilo sa vytvorit glfw okno!")
        glfwTerminate()
        return -1
    # kontext
    glfwMakeContextCurrent(glfwWin)
    # reakcia glfw okna na zmenu velkosti
    glfwSetFramebufferSizeCallback(glfwWin, glfwWin_framebuffersize_callback)
    # reakcia glfw okna na pohyb mysi
    glfwSetCursorPosCallback(glfwWin, glfwWin_cursorpos_callback)
    # reakcia glfw okna na pohyb kolieska mysi
    glfwSetScrollCallback(glfwWin, glfwWin_scroll_callback)
    # skrytie a zachytavanie kurzoru mysi
    glfwSetInputMode(glfwWin, GLFW_CURSOR, GLFW_CURSOR_DISABLED)

    # OpenGL stavy
    glClearColor(0.2, 0.2, 0.2, 1.0)  # farba, ktorou sa bude resetovat zasobnik farieb
    glClearDepth(1.0)  # zasobnik hlbky bude vymazavany nastavenim na 1.0
    glDepthFunc(GL_LESS)  # typ porovnavacej funkcie pre urcenie Z-vzdialenosti
    glEnable(GL_DEPTH_TEST)  # testovanie Z-vzdialenosti

    glDisable(GL_CULL_FACE)  # poradie, v ktorom budu definovane vrcholy objektov
    glFrontFace(GL_CCW)  # reprezentovat privratene, resp. odvratene plochy,
    glCullFace(GL_BACK)  # CCW - proti smeru hod. ruciciek, CW - v smere hod. ruciciek
    glEnable(GL_CULL_FACE)  # odvratene plochy sa nebudu spracovavat

    # nahratie standardnej (default) textury
    tex_id = 0
    try:
        img = Image.open("data/default.png")
        bands_count = len(img.getbands())
        tex_format = GL_RED if bands_count == 1 else \
            GL_RGB if bands_count == 3 else \
                GL_RGBA
        tex_id = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, tex_id)
        glTexImage2D(GL_TEXTURE_2D, 0, tex_format, img.width, img.height,
                     0, tex_format, GL_UNSIGNED_BYTE, img.tobytes())
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
        glGenerateMipmap(GL_TEXTURE_2D)
        img.close()
    except FileNotFoundError:
        print("Chyba nahratia standardnej (default) textury!")

    # priprava objektov - kocka, gula, valec, ihlan, plochy
    myObj: myObjects = myObjects()
    myObj.fill_arrays()
    vertnum = myObj.get_vertices_num()

    vertices = glm.array(myObj.get_vertices())
    colors = glm.array(myObj.get_colors())
    normals = glm.array(myObj.get_normals())
    texcoords = glm.array(myObj.get_texcoords())

    # Vertex Array Object
    VAO = glGenVertexArrays(1)
    glBindVertexArray(VAO)
    # v_a Vertex Buffer object pre vrcholy
    VBO1 = glGenBuffers(1)
    glBindBuffer(GL_ARRAY_BUFFER, VBO1)
    glBufferData(GL_ARRAY_BUFFER, vertices.nbytes, vertices.ptr, GL_STATIC_DRAW)
    glEnableVertexAttribArray(0)  # povolenie pola atributov
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * glm.sizeof(glm.float32), ctypes.c_void_p(0))
    # c_a Vertex Buffer object pre farby
    VBO2 = glGenBuffers(1)
    glBindBuffer(GL_ARRAY_BUFFER, VBO2)
    glBufferData(GL_ARRAY_BUFFER, colors.nbytes, colors.ptr, GL_STATIC_DRAW)
    glEnableVertexAttribArray(1)  # povolenie pola atributov
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 3 * glm.sizeof(glm.float32), ctypes.c_void_p(0))
    # n_a Vertex Buffer object pre normaly
    VBO3 = glGenBuffers(1)
    glBindBuffer(GL_ARRAY_BUFFER, VBO3)
    glBufferData(GL_ARRAY_BUFFER, normals.nbytes, normals.ptr, GL_STATIC_DRAW)
    glEnableVertexAttribArray(4)  # povolenie pola atributov
    glVertexAttribPointer(4, 3, GL_FLOAT, GL_FALSE, 3 * glm.sizeof(glm.float32), ctypes.c_void_p(0))
    # t_a Vertex Buffer object pre texturovacie suradnice
    VBO4 = glGenBuffers(1)
    glBindBuffer(GL_ARRAY_BUFFER, VBO4)
    glBufferData(GL_ARRAY_BUFFER, texcoords.nbytes, texcoords.ptr, GL_STATIC_DRAW)
    glEnableVertexAttribArray(8)  # povolenie pola atributov
    glVertexAttribPointer(8, 2, GL_FLOAT, GL_FALSE, 2 * glm.sizeof(glm.float32), ctypes.c_void_p(0))

    # priprava shader programu
    base_shader = Shader("shaders/base.vs", "shaders/base.fs")
    # pouzitie shader programu
    base_shader.use()
    # glBindAttribLocation(base_shader.program, 0, "in_pos") nemusi sa vykonat
    # glBindAttribLocation(base_shader.program, 1, "in_col") nemusi sa vykonat
    # glBindAttribLocation(base_shader.program, 4, "in_nor") nemusi sa vykonat
    # glBindAttribLocation(base_shader.program, 8, "in_uv")  nemusi sa vykonat
    base_shader.set_U1i("base_tex_unit", 0)  # zakladna texturovacia jednotka

    # hlavny programovy cyklus
    while not glfwWindowShouldClose(glfwWin):
        # casovac
        aTime = glfwGetTime()
        dTime = aTime - lTime
        lTime = aTime

        # reakcia glfw okna na stlacanie klaves = aktualizacia parametrov Avatara
        glfwWin_get_keys(glfwWin)
        projmat = myAvatar.get_proj_mat()
        mvmat = myAvatar.get_mv_mat()
        base_shader.set_UMatrix4fv("projection_mat", projmat)
        base_shader.set_UMatrix4fv("modelview_mat", mvmat)

        # zmazanie zasobnikov
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        # vykreslenie objektov
        glBindTexture(GL_TEXTURE_2D, tex_id)
        glDrawArrays(GL_TRIANGLES, 0, vertnum)

        glfwSwapBuffers(glfwWin)
        glfwPollEvents()

    # zrusenie program objektu
    glDeleteProgram(base_shader.program)
    # zmazanie VBO
    glDeleteBuffers(1, (VBO4,))
    glDeleteBuffers(1, (VBO3,))
    glDeleteBuffers(1, (VBO2,))
    glDeleteBuffers(1, (VBO1,))
    # zmazanie VAO
    glDeleteVertexArrays(1, (VAO,))
    # zmazanie textury
    glDeleteTextures(1, (tex_id,))

    glfwTerminate()

    return 0


# reakcia glfw okna na zmenu velkosti --------------------------------------------
def glfwWin_framebuffersize_callback(win, width: int, height: int) -> None:
    glViewport(0, 0, width, height)


# reakcia glfw okna na pohyb mysi ------------------------------------------------
def glfwWin_cursorpos_callback(win, xpos: float, ypos: float) -> None:
    global mouse_lastX, mouse_lastY, mouse_start

    if mouse_start:
        mouse_lastX = xpos
        mouse_lastY = ypos
        mouse_start = False

    xoffset = xpos - mouse_lastX
    yoffset = mouse_lastY - ypos

    mouse_lastX = xpos
    mouse_lastY = ypos

    myAvatar.setrot(xoffset, yoffset)


# reakcia glfw okna na pohyb kolieska mysi ---------------------------------------
def glfwWin_scroll_callback(win, xoffset: float, yoffset: float) -> None:
    myAvatar.setFOV(yoffset)


# reakcia glfw okna na stlacanie klaves ------------------------------------------
def glfwWin_get_keys(win) -> None:
    if glfwGetKey(win, GLFW_KEY_ESCAPE) == GLFW_PRESS:
        glfwSetWindowShouldClose(win, True)
    if glfwGetKey(win, GLFW_KEY_W) == GLFW_PRESS:
        myAvatar.setpos(goForward, dTime)
    if glfwGetKey(win, GLFW_KEY_S) == GLFW_PRESS:
        myAvatar.setpos(goBack, dTime)
    if glfwGetKey(win, GLFW_KEY_A) == GLFW_PRESS:
        myAvatar.setpos(goLeft, dTime)
    if glfwGetKey(win, GLFW_KEY_D) == GLFW_PRESS:
        myAvatar.setpos(goRight, dTime)


# hlavna funkcia aplikacie -------------------------------------------------------
if __name__ == "__main__":
    main()
